/**
 * An example C multithreaded program.
 *
 * WARNING: Contains a subtle threading error which can cause a deadlock!
 *          Can you find and fix it?
 *
 * by Ron Mak
 * Department of Computer Engineering
 * San Jose State University
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include "synchronize.h"

#define STUDENT_ID_BASE 101

#define CHAIR_COUNT 3
#define STUDENT_COUNT 25

#define MAX_MEETING_DURATION 5
#define OFFICE_HOUR_DURATION 60

#define STUDENT_AVAILABLE_SEMAPHORE_NAME "/student_available"

// Shared data
int chairs[CHAIR_COUNT];
int in = 0, out = 0;
int wait_count = 0;

// Mutexes
pthread_mutex_t chair_mutex;  // mutex protects chairs and wait count
pthread_mutex_t print_mutex;  // mutex protects printing

// Semaphore
sem_t *student_available;     // professor waits on this semaphore

// Professor's office hour timer
struct itimerval professor_timer;
time_t startTime;

int meeting_id = 0;
int arrivals_count = 0;
int leaves_count = 0;
int meetings_count = 0;
int parfore_count = 0;

// Functions to run in the threads
void *professor(void *param);
void *student(void *param);

// Critical regions
void professor_meets_student();
void student_arrives(int id);
void print(char *event);

bool times_up = false;  // is office hour is over?

// Timer signal handler.
void timer_handler(int signal);

int main(int argc, char *argv[])
{
    int student_ids[STUDENT_COUNT];
    int professor_id = 0;

    // Initialize the mutexes locks.
    INIT(&chair_mutex, "chair_mutex");
    INIT(&print_mutex, "print_mutex");

    // Remove the old semaphore and create a new one.
    REMOVE(STUDENT_AVAILABLE_SEMAPHORE_NAME);
    student_available = OPEN(STUDENT_AVAILABLE_SEMAPHORE_NAME, 0);

    srand(time(0));
    printf("TIME | MEETING | WAITING     | EVENT\n");

    // Create the professor thread.
    pthread_t professor_thread_id;
    pthread_attr_t profAttr;
    pthread_attr_init(&profAttr);
    pthread_create(&professor_thread_id, &profAttr, professor, &professor_id);

    // Create the student threads.
    for (int i = 0; i < STUDENT_COUNT; i++)
    {
        student_ids[i] = STUDENT_ID_BASE + i;
        pthread_t studentThreadId;
        pthread_attr_t studentAttr;
        pthread_attr_init(&studentAttr);
        pthread_create(&studentThreadId, &studentAttr, student, &student_ids[i]);
    }

    // Set the timer signal handler.
    signal(SIGALRM, timer_handler);

    // Wait for the professor to complete the office hour.
    pthread_join(professor_thread_id, NULL);

    // Remaining waiting students leave.
    meeting_id = 0;
    while (wait_count-- > 0)
    {
        int studentId = chairs[out];
        out = (out + 1)%CHAIR_COUNT;
        leaves_count++;

        char event[80];
        sprintf(event, "Student %d leaves",  studentId);
        print(event);
    }

    // Final statistics.
    printf("\n");
    printf("%5d students arrived\n", arrivals_count);
    printf("%5d students met with Prof. Fore\n", meetings_count);
    printf("%5d students left without meeting\n", leaves_count);
    printf("%5d times Prof. Fore worked on ParFore\n", parfore_count);

    return 0;
}

// The professor thread function.
void *professor(void *param)
{
    time(&startTime);
    print("Professor opens her door");

    // Set the timer for for office hour duration.
    professor_timer.it_value.tv_sec = OFFICE_HOUR_DURATION;
    setitimer(ITIMER_REAL, &professor_timer, NULL);

    // Meet students until the office hour is over.
    do
    {
        professor_meets_student();
    } while (!times_up);

    print("Professor closes her door");
    return NULL;
}

// The professor meets a student (or works on ParFore).
void professor_meets_student()
{
    // No student waiting, so work on ParFore language.
    if (wait_count == 0)
    {
        print("Professor works on ParFore");
        parfore_count++;
    }

    if (!times_up)
    {
        WAIT(student_available, "student_available");

        LOCK(&chair_mutex, "chair_mutex");
        {
            meeting_id = chairs[out];
            out = (out+1)%CHAIR_COUNT;
            wait_count--;
        }
        UNLOCK(&chair_mutex, "chair_mutex");

        char event[80];
        sprintf(event, "Professor meets with student %d",  meeting_id);
        print(event);

        // Meet with the student.
        sleep(rand()%MAX_MEETING_DURATION + 1);
        meetings_count++;

        sprintf(event, "Professor finishes with student %d",  meeting_id);
        meeting_id = 0;
        print(event);
    }
}

// The student thread function.
void *student(void *param)
{
    int id = *((int *) param);

    // Students will arrive at random times during the office hour.
    int duration = rand()%OFFICE_HOUR_DURATION;
    sleep(duration);
    student_arrives(id);

    return NULL;
}

// A student arrives.
void student_arrives(int id)
{
    char event[80];
    arrivals_count++;

    if (wait_count < CHAIR_COUNT)
    {
        // Seat a student into a chair.
        LOCK(&chair_mutex, "chair_mutex");
        {
            chairs[in] = id;
            in = (in+1)%CHAIR_COUNT;
            wait_count++;
        }
        UNLOCK(&chair_mutex, "chair_mutex");

        sprintf(event, "Student %d arrives and waits", id);
        print(event);

        SIGNAL(student_available, "student_available");
    }
    else
    {
        leaves_count++;
        sprintf(event, "Student %d arrives and leaves", id);
        print(event);
    }
}

void timer_handler(int signal)
{
    times_up = true;  // office hour is over
}

void print(char *event)
{
    // Print a line for each event:
    //   elapsed time
    //   who is meeting with the professor
    //   who is waiting in the chairs
    //   what event occurred
    time_t now;
    time(&now);

    double elapsed = difftime(now, startTime);
    int min = 0;
    int sec = (int) elapsed;

    if (sec >= 60)
    {
        min++;
        sec -= 60;
    }

    LOCK(&print_mutex, "print_mutex");
    {
        // Elapsed time.
        printf("%1d:%02d | ", min, sec);

        // Who's meeting with the professor.
        if (meeting_id > 0) printf("%5d   |", meeting_id);
        else                printf("        |");

        int k = 0;

        LOCK(&chair_mutex, "chair_mutex");
        {
            int i = out;
            int j = wait_count;

            // Who's waiting in the chairs.
            while (j-- > 0)
            {
                printf("%4d", chairs[i]);
                i = (i + 1)%CHAIR_COUNT;
                k++;
            }
        }
        UNLOCK(&chair_mutex, "chair_mutex");

        // What event occurred.
        while (k++ < CHAIR_COUNT) printf("    ");
        printf(" | %s\n", event);
    }
    UNLOCK(&print_mutex, "print_mutex");
}
